//disc_burn
// This program will exercise the hard drive.  The test uses a random number 
// generator to seek to various areas of the disk within a file and
// then writes a character string to that location. When exiting, either by 
// user intervention (pressing a key during the test) or by time, the program
// will remove the burn-in file from the hard drive.
// Specifying no options will put no time limit on the test and will dynamically
// create the file size according to the available space left on the disk (up
// to a maximum of 200MB).

// Three command line options are allowed: -s : specifies the file size, valid 
// range is from 1MB to 2000MB; and -t : specifies the time to run the program,
// valid range is from 1 min. to 360 min., and -h : displays command line usage
// information and exits(overrides any other options).
// syntax: disc_burn [-h] [-t time] [-s filesize]
// 	time is an integer between 1 and 360
// 	filesize is an integer between 1 and 2000

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#define BURN_FILE "burn_in"	//burn file name
#define BUF_SIZE 26		//buffer size of buf

//subtracted off space available when detecting filesize allowed
#define SIZE_BUFFER 20		 

int main(int argc, char *argv[])
{
    int fd, fd1;	//file descriptors
    char size[50];	//holds filesize data returned from shell command: df

    //This is what we are writing at each location
    char buf[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    
    off_t offset;	//offset in bytes into file
    int random, option;
    pid_t pid;
    long int elapsed_t, elapsed_min, filesize, burntime = 0;
    time_t timesec;
    int s;		//stores the space available
            
    //used by the child process for key detection
    int in1;
    int c;
    struct termios attr;
    
    //detect space available and adjust file size accordingly
    system("df disc_burn | grep da | gawk '{print $4}' > sizedf");
    fd1 = open("sizedf", O_RDONLY);
    s = read(fd1, size, 50);
    close(fd1);
    if (s == 0)
    {
	printf("error detecting space available: unreadable or no space left");
    }
    filesize = ((long int)atoi(size))/1000 - SIZE_BUFFER;
    if (filesize > 200)
    {
	filesize = 200;
    }
    if (filesize < 1)
    {
	printf("Not enough space left on harddrive\n");
	exit(1);
    }
    if (remove("sizedf") == -1)
    {
	printf("unable to remove sizedf\n");
	exit(1);
    }
    
    //get any options specified
    while ( (option = getopt(argc, argv, "ht:s:")) != EOF)
    {
	switch (option)
	{
	case 'h':
	    printf("usage: disc_burn [-h] [-t time] [-s filesize];\n");
	    printf("where time is in minutes and filesize is in MB\n");
	    printf("(1 <= time <= 360) (1 <= filesize <= 2000)\n");
	    printf("h: displays this usage information\n\n");
	    exit(0);
	case 't':
	    burntime = (long int)atoi(optarg);
	    //check to make sure it is between 1min and 360min (max. 6hrs)
	    if ((burntime < 1) || (burntime > 360))
	    {
		printf("usage: disc_burn [-h] [-t time] [-s filesize];\n");
		printf("where time is in minutes and filesize is in MB\n");
		printf("(1 <= time <= 360) (1 <= filesize <= 2000)\n");
		printf("h: displays this usage information\n\n");
		exit(1);
	    }
	    break;	
	case 's':    
	    filesize = (long int)atoi(optarg);
	    //check to make sure it is a valid file size (1MB to 2000MB)
	    if ((filesize < 1) || (filesize > 2000))
	    {
		printf("usage: disc_burn [-h] [-t time] [-s filesize];\n");
		printf("where time is in minutes and filesize is in MB\n");
		printf("(1 <= time <= 360) (1 <= filesize <= 2000)\n");
		printf("h: displays this usage information\n\n");
		exit(1);
	    }
	    break;
	case ':':
	    exit(1);
	case '?':
	    exit(1);
	default:
	    printf("usage: disc_burn [-h] [-t time] [-s filesize];\n");
	    printf("where time is in minutes and filesize is in MB\n");
	    printf("(1 <= time <= 360) (1 <= filesize <= 2000)\n");
	    printf("h: displays this usage information\n\n");
	    exit(1);
	}
    }
        
    //adjust to bytes and subtract the size to the buffer being written
    filesize = (filesize * 1000000) - BUF_SIZE;
    
    //initialize all time values
    elapsed_t = 0;
    elapsed_min = 0;
    timesec = time(NULL);
    
    printf("\n\nNetWinder HardDrive Burn-In\n\n");
    
    //create the file
    if (( fd = open(BURN_FILE, O_RDWR | O_CREAT | O_TRUNC, 
	S_IRUSR | S_IWUSR )) == -1)
    {
	printf("open error\n");
    }
    //write to the beginning of the file
    if (write(fd, buf, BUF_SIZE) != BUF_SIZE)
    {
	printf("write error\n");
	perror(argv[0]);
    }
    //skip to the end of the file
    if (lseek(fd, filesize, SEEK_SET) == -1)
    {
	printf("seek error\n");
	perror(argv[0]);
    }
    //write to the end of the file
    if (write(fd, buf, BUF_SIZE) != BUF_SIZE)
    {
	printf("write error\n");
	perror(argv[0]);
    }
        
    if ( ( pid = fork()) < 0)
    {
	printf("error forking\n");
    }
    else if (pid != 0) 			//parent
    {
	printf("burning...\n");
	printf("press any key to cancel burn-in\n");
	//if the test is not time limited then display time status
	if ( burntime != 0 )
	{
	    printf("\nTime elapsed = %.3d min\tTime remaining = %.3d min",
		     elapsed_min, (burntime - elapsed_min));
	    fflush(NULL);
	}
	for (;;)
	{
	    //update time values and display every 60 seconds
	    if ( burntime != 0 )
	    {
		elapsed_t = time(NULL) - timesec;
		if (elapsed_min < (elapsed_t / 60))
		{
		    elapsed_min = elapsed_t / 60;
		    printf("\rTime elapsed = %.3d min\tTime remaining = %.3d min",
		             elapsed_min, (burntime - elapsed_min));
		    fflush(NULL);
		}
		//cleanup and exit when burntime has been reached
		if ( elapsed_t >= (burntime * 60))
		{
		    close(fd);
		    printf("\nperforming cleanup...\n");
		    kill(pid, SIGTERM);		//kill child
		    if (remove("burn_in") == -1)
		    {
			printf("unable to remove the burn-in file\n");
			exit(1);
		    }
		    exit(0);
		}
	    }
	    
	    //generate random number for seek
	    random = (int)((double)filesize * rand() / RAND_MAX);
	    if (lseek(fd, random, SEEK_SET) == -1)
	    {
		printf("seek error\n");
		perror(argv[0]);
	    }
	    if (write(fd, buf, BUF_SIZE) != BUF_SIZE)
	    {
		printf("write error\n");
		close(fd);
		if (errno == ENOSPC)
		{
		    perror(argv[0]);
		    if (remove("burn_in") == -1)
		    {
			printf("unable to remove the burn-in file\n");
			exit(1);
		    }
		    exit(0);
		}
	    }
	    //checks to see if child has exited,if so...cleanup and exit
	    if (waitpid(pid, NULL, WNOHANG) != 0)
	    {
		printf("\nperforming cleanup...\n");
		close(fd);
		if (remove("burn_in") == -1)
		{
		    printf("unable to remove the burn-in file");
		    exit(1);
		}
		exit(0);
	    }
	}	
    }
    else				//child
    {   
	//detects when a key has been pressed then kills itself
	
	in1 = fileno(stdin);
	tcgetattr(in1, &attr);
	attr.c_lflag &= ~(ICANON);
	attr.c_cc[VMIN] = 1;
	attr.c_cc[VTIME] = 0;
	tcsetattr(in1, TCSAFLUSH, &attr);
	read(in1, &c, 1);
	printf("\b ");
	
	exit(0);
    }
	
}	    