/* things to do: changed setup_vidcap() so that its easier to call */
 /* NOTE TO THE READER:
 I really would like to get video capture working in both
 text and graphics modes. For example, it would be nice to
 take captures for a WebCam application even if the system
 is in text mode

 Capture actually operates in text mode, but I cannot seem
 to read the capture buffer into kernel space. It seems
 folded, probably due to vga chain addressing.
 
 You will see some #ifdef EXPERIMENT
 sections below. These were my unsuccessful attempts to
 set the vga into linear mode so I could read the
 capture buffer.

 If one of you succeeds in fixing this, please let me know
 patb@corelcomputer.com
 */
 

/**
 * Author: Milind Changire
 *
 * Copyright Corel Comptuer Corp (c) 1997-1998
 *
 * Video Capture driver for the Philips SAA7111 interfacing via reg 3CF/B6 
 * of the CyberPro2000 using the I2C interface.
 * The IGS SDK for video capture has been used to develop this device driver.
 * http://www.igst.com
 *
 *
 * Note:
 *   Please consult Pat Beirne <patb@corelcomputer.com> or 
 *   San Mehat <sanm@corelcomputer.com> about Linux Device Driver Development.
 *   If you are a novice to Linux Device Driver Development then you must read
 *   and use
 *   "Writing Linux Device Drivers"
 *   (http://www.redhat.com/~johnsonm/devices.html)
 *
 *                 and
 *
 *   "Linux Kernel Hacker's Guide"
 *   (http://www.redhat.com:8080/HyperNews/get/khg.html)
 *
 * Note:
 *   All operations handle YUV422 data and NOT RGB data.
 *   The YUV to RGB conversion routines can be used to convert a captured frame
 *   to create a photograph or portrait for a Video Conferencing application.
 *
 * Most routines return 0 on success, <0 on error
 *
 * Major revision: Pat Beirne Sept/1998
 *    -new ioctls
 *    -no more screen painting
 *    -data moves go through read/write
 *    -minor device numbers to separate main overlay from 2nd overlay
 *    -backdoor call for Xserver and svgalib: to set capture address
 *    -capture area moved to high vga ram (instead of low vga ram)
 *    -ioctl for controlling the 7111 (brightness & contrast)
 *    -separate ioctls for capture and display
 *    -most setting can be read, so that apps can share the device
 *    -man page
 *    -simplified the internal state control
 *    -compatible with Video4Linux2
 * Minor: Oct 15/98
 *    -allow open() even if not in graphics mode
 *    -_S_FMT, _S_WIN will fail if not in graphics mode
 *
 *
 * HINT: vafc is the video-feature-connector through which the
 * Philips chip gives us the camera data
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <asm/segment.h>

/*****************************************************************************/
#include "videodev.h"
#include "igs.h"
#include "igsvideo.h"
//#include "fifotab.h"
#include "vidcap.h"

/* all our IO is w.r.t. this memory address */
unsigned char* pIO = (unsigned char*) 0xD0800000;

/* state of the capture-display driver; big */
static struct nw_cap_device dev;

/* state of the overlay driver */
//static struct _state overlay2_state;

static const char vidcap_name[] = "vidcap";
static int vidcap_major = VIDEO_MAJOR;

/* the device driver entry points */
static int  open_vidcap (struct inode *inodep, struct file *filep);
static void release_vidcap (struct inode *inodep, struct file *filep);
static int  vidcap_ioctl (struct inode *inodep, struct file *filep,
                          unsigned int cmd, unsigned long arg);
static int  vidcap_read(struct inode* inode,struct file* file,
                        char* buf,int count);
static int  vidcap_write(struct inode* inode,struct file* file,
                         const char* buf,int count);

static void release_video ();
static int  setup_vidcap ();

static struct file_operations vidcap_fops =
{
  NULL,							/* no special lseek              */
  vidcap_read, 						/* no special read               */
  vidcap_write,						/* no special write              */
  NULL,							/* no special readdir            */
  NULL,							/* no special select             */
  vidcap_ioctl,
  NULL,							/* no special mmap               */
  open_vidcap,
  release_vidcap,
  NULL,							/* no special fsync              */
  NULL,							/* no special fasync             */
  NULL,							/* no special check_media_change */
  NULL							/* no special revaldate          */
};

/* some capture parameters */
#define MAX_WIDTH   704
#define MAX_HEIGHT  PAL_SRC_HEIGHT /* about 572 */
#define MIN_WIDTH    32
#define MIN_HEIGHT   32
#define MAX_CAPTURE_BUFFERS 1

/* return from VIDIOC_G_CAP */
static struct video_capability this_device =
{
    "NetWinder Video Capture",
    VID_TYPE_CAPTURE | VID_TYPE_PREVIEW,
    1,0,       /* channels */
    MAX_WIDTH, MAX_HEIGHT,
    MIN_WIDTH, MIN_HEIGHT,
    30 /* max frame rate; HA! */
};


/* capture formats;  for VIDIOC_G_CAPFMT & VIDIOC_G_FBUFFMT     */
struct video_fmtdesc capfmt[] = 
{
	{	0, {"YUUV"},
		PIX_FMT_YUYV,  16, 0, {0, 0},
	},
	{	1, {"RGB-24"},
		PIX_FMT_RGB24,   24, 0, {0, 0},
	},
	{	2, {"RGB-16 (5-6-5)"},
		PIX_FMT_RGB565,  16, 0, {0, 0},
        }
};
#define NUM_CAPFMT sizeof(capfmt)/sizeof(struct video_fmtdesc)

/* controls (analog parameters); for VIDIOC_QUERYCTRL */
struct video_queryctrl video_control[] =
{
    /* index, name, low_val, high_val, incr, default */
    { CID_BRIGHTNESS, "brightness", 0, 0xff, 1, 0x80},
    { CID_CONTRAST,   "contrast",   0, 0x7f, 1, 0x40},
    { CID_SATURATION, "saturation", 0, 0x7f, 1, 0x47},
    { CID_HUE,        "hue",        -0x80, 0x7f, 1, 0}
};


/* most of these globals are from an old design;
 I suppose they really should move into the dev structure */
#define VGA_IO_ADDR 0xd0000000
unsigned long dwChipID = 0;             /* from the CyberPro */
int wMPEG2Enable = 0;			/* flag to inicate whether full 
					   * screen capture is supported or
					   * not 
					   */




/**************************************************************************
 ****************** init code *********************************************
 **************************************************************************/
/**
 * int setup_vidcap()
 * called during device open
 * refreshes current vga information
 * assumes that Philips SAA7111 is already set up
 *
 * Returns: -ENXIO if Philips SAA7111 can't be detected or
 *          if current video mode does not support Video Capture.
 */
static int
setup_vidcap ()
{
#ifndef EXPERIMENT
    if (IGS_IsGraphics() == FALSE) {
        printk("vidcap.o: Error, the display must be in graphics mode for capture services\n");
        return -ENXIO;
    }
#endif
    // turn on the 7111 hardware
    set_CPLD(0,4);

    if (DetectDecoder (&dev.camera_data_size) == PHILIPS7111)
        InitPhil7111 ();
    else
        return -ENXIO;

    /* query the VGA for some current settings;
     sets up dev.screen with the current vga bytes_per_pixel,
     width, height, etc, etc */
    IGS_GetCurrentScreen(&dev);
    dev.screen.display_memory_used =
        dev.screen.x_ext * dev.screen.y_ext * dev.screen.bytes_per_pixel;
    dev.screen.total_video_memory = IGS_GetVMemSize();
    /* it is too early to guess at a capture area, but
     to be safe, put it after the display area */
    if (!dev.mem_buf_set)
        dev.screen.offset_vidcap_mem = dev.screen.display_memory_used;
#ifdef DEBUG
    printk("vidcap: setup\n"
           " visible screen w/h = %d/%d\n left/top = %d/%d\n"
           " extents w/h = %d/%d\n memory used = %d\n"
           " on-board memory = %d\n bytes per pixel = %d\n",
           dev.screen.width, dev.screen.height, dev.screen.x_orig, dev.screen.y_orig,
           dev.screen.x_ext, dev.screen.y_ext, dev.screen.display_memory_used,
           dev.screen.total_video_memory, dev.screen.bytes_per_pixel);
#endif

  /* calculate the FIFO setting for this graphics mode */
  if (!GetVideoFIFO(&dev))
  {
     printk ("vidcap.c: Error, the current mode does not support capture!\n");
     return -ENXIO;			/* error */
  }
      

  IGS_InitVideo (&dev);

  dev.video_initialized = TRUE;
  printk("vidcap: video_initialized set to %d\n",dev.video_initialized);
  
  return 0;
}



/* try to set a new capture format
 in: struct video_format* requested video format;
           may be changed by this code;
           modified version deposited in dev.capture_format
 struct video_parm*   extra qualifiers on the capture
           such as xflip, y flip and hi_res
 
 first, update the device structures, checking along the way;
 then, forward these changes to the hardware (SetCaptureParms())
 
 if success, update this_format
 if error, return error
 */
/* NOT YET DONE !!!! use xMirror, yMirror and hi_res bits */
int capture_new_format(struct video_format* fmt,
                      struct video_parm* parm)
{
#ifdef DEBUG
    printk("capture_new_format w/h %d/%d, pf %x\n",fmt->width,fmt->height,fmt->pixelformat);
#endif
    if (fmt->width>MAX_WIDTH || fmt->height>MAX_HEIGHT)
        return -EINVAL;
    switch(fmt->pixelformat) {
    case PIX_FMT_YUYV:
    case PIX_FMT_RGB565:
        dev.capture_format.pixelformat = fmt->pixelformat;
        dev.capture_format.depth = 16;
        break;
    case PIX_FMT_RGB24:
        dev.capture_format.pixelformat = fmt->pixelformat;
        dev.capture_format.depth = 24;
        break;
    default:
        return -EINVAL;
    }
//    printk("capture_new_format, past pixelformat set\n");
    /* all the other input params ignored */
    dev.capture_format.width = fmt->width & ~1;
    if (dev.capture_format.width<MIN_WIDTH)
        dev.capture_format.width = MIN_WIDTH;
    dev.capture_format.height = fmt->height & ~1;
    if (dev.capture_format.height<MIN_HEIGHT)
        dev.capture_format.height = MIN_HEIGHT;
    
    dev.capture_format.flags = FMT_FLAG_BYTESPERLINE |
        (dev.capture_format.height>240 ? FMT_FLAG_INTERLACED : 0) |
        (dev.capture_format.pixelformat==PIX_FMT_YUYV ? FMT_CS_601YUV : 0);
    dev.capture_format.bytesperline = dev.capture_format.width * dev.capture_format.depth / 8;
    dev.capture_format.sizeimage = dev.capture_format.bytesperline * dev.capture_format.height;

    /* try to set up the hardware;
     give SetupAndCapture() a chance to return an error */
    /* NOTE: this is different than the WINNOV driver;
     we start capture right away; they wait till a read() cmd */
    return SetupAndCapture(&dev.capture_format, parm);
}

/* continuation of the above, except does not modify the dev structure
 this one actually sets the hardware regs
 */
int
SetupAndCapture(struct video_format * fmt,
               struct video_parm* parm) {
    /* do some minor checking on fmts */
    if (fmt->width & 1)
        return -EINVAL;
    if ((fmt->width > MAXMPEG1) && (wMPEG2Enable == 0))
        return -EINVAL;
    /* !! this would be a good place to check if there is
     * enough video ram to do a capture
     * if we know how much video ram has been used !!*/
#ifdef DEBUG
    printk("screen info: w/h= %d/%d, memused=%d, totalmem=%d\n",
           dev.screen.x_ext, dev.screen.y_ext, dev.screen.display_memory_used,
           dev.screen.total_video_memory);
    printk("capture request: w/h= %d/%d\n",fmt->width,fmt->height);
#endif
    if (fmt->width * fmt->height * 2
         + dev.screen.display_memory_used
         + 2048 /* for the hardware cursor and ping pong buffer */
         > dev.screen.total_video_memory)
        return -EINVAL;
    
    /* set the start address for capture */
    if (!dev.mem_buf_set) {
        dev.screen.offset_vidcap_mem = dev.screen.total_video_memory
            - 2048
            - (dev.capture_format.width * dev.capture_format.height * 2);
    }
    /* set the scaling */
    IGS_SetCaptureScaling(&dev);

#ifdef DEBUG
    printk("set capture address to %x\n",dev.screen.offset_vidcap_mem);
#endif
    IGS_SetCaptureAddr(&dev,dev.screen.offset_vidcap_mem);

    /* turn on the capture; does an Unfreeze() */
    IGS_CaptureOn();
    dev.capturing = TRUE;
    return 0;
}



/****************************************************
 * Device Driver exported functions.
 ****************************************************/
static int
open_vidcap (struct inode *inodep, struct file *filep)
{
    int r;

    if (MINOR(inodep->i_rdev)==VIDCAP_MINOR_CAPTURE) {
        /* initialize dev; it is mostly 0, except for these */
        if (0==dev.open_count) {
            /* there is only one thing we must preserve
             across opens: the mem buff address */
            int tmp = 0;
            if (dev.mem_buf_set)
                tmp = dev.screen.offset_vidcap_mem;
            memset(&dev,0,sizeof(dev));
            dev.control[CONTROL_BRIGHTNESS] = 0x80;
            dev.control[CONTROL_CONTRAST] = 0x40;
            dev.control[CONTROL_SATURATION] = 0x40;
            dev.control[CONTROL_HUE] = 0;
            dev.videoStandard = NTSC;
            dev.uses_chromakey = 1; /* and color == black */
            if (tmp) {
                dev.screen.offset_vidcap_mem = tmp;
                dev.mem_buf_set = TRUE;
            }
        }
        dev.open_count++;
        printk("vidcap: after open count %d\n",dev.open_count);
        MOD_INC_USE_COUNT;
    }
    else
        printk("vidcap: unknown minor number WRITE MY CODE\n");
    return 0;
}

static void
release_vidcap (struct inode *inodep, struct file *filep)
{
    /* if this is the last release, turn off the power-hungry 7111 */
    if (--dev.open_count == 0)
        release_video();
     printk("vidcap: after close count %d\n",dev.open_count);
    
    MOD_DEC_USE_COUNT;
}


#ifdef EXPERIMENT
static unsigned char regs[98] = {
  0x5F,0x4F,0x52,0x9F,0x53,0x1F,0x0B,0x3E,	/* CR00-CR17 */
  0x00,0x40,0x00,0x00,0x00,0x28,0x00,0x00,
  0xEB,0x2D,0xDF,0xC8,0x00,0xEB,0xEE,0xC3,
  0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,	/* AR00-AR14 */
  0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,
  0x01,0xFF,0x0F,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x60,0x05,0x0F,	/* GR00-GR08 */
  0xFF,
  0x03,0x01,0x0F,0x00,0x0E,			/* SR00-SR04 */
  0xE3,						/* MISC_OUT  */
  /* Extended (count = 0x26) */
  0x10,0x10,0x08,0x00,0xC9,0x08,0x25,0x00,
  0x00,0x0D,0x06,0x0A,0x00,0xB2,0x08,0x00,
  0x03,0x09,0x00,0x00,0x00,0x0B,0x10,0x45,
  0x30,0x1B,0x1E,0x00,0x02,0x08,0x00,0xC8,
  0xD2,0xCE,0xDB,0x54,0x11,0x10,
};
#endif


/* !!! NOTE: needs to be fixed
 needs sync with the capture buffer
 non-block mode needs to return -EAGAIN
 needs to inform select()
 */
static int
vidcap_read(struct inode* inode,
            struct file* file,
            char* buf,
            int count)
{
    int s1, s2, s4, g5, g6, a10, a11, restore_reg = 0;
#ifdef DEBUG
//    printk("vidcap_read: minor number %d\n",MINOR(inode->i_rdev));
#endif
    if (MINOR(inode->i_rdev)==VIDCAP_MINOR_CAPTURE) {
        /* where are we currently */
        unsigned long p = file->f_pos;
        char* pRead;
        int bytes_per_pixel =
            (dev.capture_format.pixelformat == PIX_FMT_RGB24 ? 3 : 2);
        int mem_size = dev.capture_format.width *
            dev.capture_format.height * bytes_per_pixel;

        if (!dev.video_initialized)
            return -EIO;
        
        /* synchronous */
        if (file->f_flags & O_NONBLOCK)
            IGS_SetFreezeVidAsync(ON);
        else
            IGS_SetFreezeVid(ON);
#ifdef DEBUG
        /* assert */
        if (mem_size != dev.capture_format.sizeimage)
            printk("vidcap: sizeimage not set correctly %d %d\n",
                   mem_size, dev.capture_format.sizeimage);
        if (bytes_per_pixel != dev.capture_format.depth/8)
            printk("vidcap: cap_format.depth not set correctly %d %d\n",
                   bytes_per_pixel, dev.capture_format.depth/8);
#endif
        
        /* check on count */
        if (count + p > mem_size)
            count = mem_size - p;
        pRead = (char*)dev.screen.offset_vidcap_mem + 0xD0000000 + p;
#ifdef DEBUG
//        printk("vidcap_read: reading from %x\n",pRead);
#endif
        
#ifdef EXPERIMENT
        if (IGS_IsGraphics() == FALSE)
        {
            int i;
            for(i=0; i<6; i++)
                outw_p((regs[60+i]<<8) | i+0x10, IGSEXTINDEX);
            for(i=0; i<4; i++)
                outw_p((regs[66+i]<<8) | i+0x30, IGSEXTINDEX);
            for(i=0; i<11;i++)
                outw_p((regs[70+i]<<8) | i+0x50, IGSEXTINDEX);
            for(i=0; i<11;i++)
                outw_p((regs[81+i]<<8) | i+0x70, IGSEXTINDEX);

            /* misc */
            outb_p(0xE3,0x3cc);

            outw_p(0x0100,SEQINDEX);
            for(i=0; i<4; i++)
                outw_p((regs[i+55]<<8) | i+1, SEQINDEX);
            outw_p(0x0300,SEQINDEX);

            for(i=0; i<24; i++)
                outw_p((regs[i]<<8) | i, CRTINDEX);
            
            for(i=0; i<9; i++)
                outw_p((regs[i+45]<<8) | i, IGSEXTINDEX);
            
            for(i=0; i<5; i++) {
                inb_p(0x3da);
                outb_p(i+0x10, ATTRINDEX);
                outb_p(regs[i+40], ATTRINDEX);
            }
            

            
            outb_p(1, SEQINDEX);
            s1 = inb_p(SEQDATA);
            outb_p(1, SEQDATA);

            outb_p(2, SEQINDEX);
            s2 = inb_p(SEQDATA);
            outb_p(0xF, SEQDATA);

            outb_p(4, SEQINDEX);
            s4 = inb_p(SEQDATA);
            outb_p(0xE, SEQDATA);
            outb_p(5, IGSEXTINDEX);
            g5 = inb_p(IGSEXTDATA);
            outb_p(0x60, IGSEXTDATA);

            outb_p(6, IGSEXTINDEX);
            g6 = inb_p(IGSEXTDATA);
            outb_p(5, IGSEXTDATA);

            inb_p(0x3da);
            outb_p(0x10, ATTRINDEX);
            a10 = inb_p(ATTRINDEX+1);
            inb_p(0x3da);
            outb_p(0x10, ATTRINDEX);
            outb_p(1, ATTRINDEX);
            outw_p(0x6005,0x3ce);
            outw_p(0x0506,0x3ce);
            outw_p(0xc317,0x3d4);
            outw_p(0x0014,0x3d4);
//          outw_p(0x0957,IGSEXTINDEX);
            outw_p(0x1190,IGSEXTINDEX);
            restore_reg = 1;
        }
#endif

        if ((file->f_flags & O_NONBLOCK)
            && !IGS_IsFreezeDone())
            return -EAGAIN;
        
        switch(dev.capture_format.pixelformat) {
        case PIX_FMT_YUYV:
            /* the easy one; just copy the yuv bytes */
            if (copy_to_user(buf, pRead, count))
                return -EFAULT;
            break;
        case PIX_FMT_RGB24:
            /* it is too hard to deal with non-triplets, so return */
            if ( (p%3) != 0)
                return -EINVAL;
            ConvertYUVtoRGB24(buf, pRead,
                                  dev.capture_format.width,
                                  dev.capture_format.height);
            break;
        case PIX_FMT_RGB565:
            if (p & 1)
                return -EINVAL;
            ConvertYUVtoRGB16(buf, pRead,
                              dev.capture_format.width,
                              dev.capture_format.height);
            break;
        default:
            return -EINVAL;
        }
        IGS_SetFreezeVidAsync(OFF);
#ifdef EXPERIMENT
        if (restore_reg)
        {
            outb_p(0x10, ATTRINDEX);
            outb_p(a10, ATTRINDEX);
            outw_p((g5<<8) | 5, IGSEXTINDEX);
            outw_p((g6<<8) | 6, IGSEXTINDEX);
            outw_p((s1<<8) | 1, SEQINDEX);
            outw_p((s2<<8) | 2, SEQINDEX);
            outw_p((s4<<8) | 4, SEQINDEX);
            outw_p(0xa317,0x3d4);
            outw_p(0x1f14,0x3d4);
            outw_p(0x0857,IGSEXTINDEX);
            outw_p(0x1090,IGSEXTINDEX);
        }
#endif
        /* if the user has read beyond the limits of the memory, we auto
         return the pointer to 0 */
        p += count;
        if (p >= mem_size)
             p = 0;

        file->f_pos = p;
        return count;
    }
    else /* the other minor numbers */
        return -EINVAL;
}

int vidcap_write(struct inode* inode,
            struct file* file,
            const char* buf,
                 int count)
{
   printk("vidcap write: WRITE MY CODE!!!!\n");
}


static int
vidcap_ioctl (struct inode *inodep, struct file *filep, unsigned int cmd, unsigned long arg0)
{
    int xpos = 0, ypos = 0, i;
    void* arg = (void*)arg0;
    
    switch (cmd)
    {
    case VIDIOC_G_CAP:
    {
        struct video_capability b;
        sprintf(b.name, "NetWinder Video (%d)", MINOR(inodep->i_rdev));
        b.type = VID_TYPE_CAPTURE |
        #ifdef notyet
            VID_TYPE_SELECT |
        #endif
            VID_TYPE_PREVIEW;
        b.inputs = 1;
        b.audios = 0;
        b.maxwidth = 704;
        b.maxheight = 480;
        b.minwidth = 32;
        b.minheight = 32;
        b.maxframerate = 30;
        if (copy_to_user(arg, &b, sizeof(b)))
            return -EFAULT;
        return 0;
    }

    case VIDIOC_G_FBUFFMT:
    case VIDIOC_G_CAPFMT:
    {
        struct video_fmtdesc f;
        if (copy_from_user(&f, arg, sizeof(f)))
            return -EFAULT;
        if (f.index < 0 || f.index >= NUM_CAPFMT)
            return -EINVAL;
        f = capfmt[f.index];
        if (copy_to_user(arg, &f, sizeof(f)))
            return -EFAULT;
        return 0;
    }

    case VIDIOC_G_FMT:
        {
            if (copy_to_user(arg, &dev.capture_format, sizeof(dev.capture_format)))
                return -EFAULT;
            return 0;
        }

    case VIDIOC_S_FMT:
        {
            struct video_format fmt;
            int r;
            if (copy_from_user(&fmt, arg, sizeof(fmt)))
                return -EFAULT;
            if (!dev.video_initialized) {
                r = setup_vidcap();
                if (r)
                    return r;
            }
            /* try this setting */
            r = capture_new_format(&fmt, &dev.capture_param);
            /* if not ok, return error */
            if (r<0)
                return r;
            /* inform caller of any modification to his request */
            copy_to_user(arg, &dev.capture_format, sizeof(dev.capture_format));
            return 0;
        }

    case VIDIOC_REQBUFS:
        return -EINVAL;
        
#ifdef notyet
        {
            struct video_requestbuffers req;
            if (copy_from_user(&req, arg, sizeof(req)))
                    return -EFAULT;
            if (!mmap_request_buffers(dev, &req))
                    return -EINVAL;
            if (copy_to_user(arg, &req, sizeof(req)))
                    return -EFAULT;
            return 0;
        }

    case VIDIOC_QUERYBUF:
    {
            struct video_buffer buf;
            int	i;
            if (o->noncapturing)
                    return -EPERM;
            if (copy_from_user(&buf, arg, sizeof(buf)))
                    return -EFAULT;
            i = buf.index;
            if (i >= MAX_CAPTURE_BUFFERS ||
                !dev->stream_buf[i].requested ||
                (buf.type & BUF_TYPE_field) != 
                 (dev->stream_buf[i].vidbuf.type & BUF_TYPE_field))
                    return -EINVAL;
            buf = dev->stream_buf[i].vidbuf;
            if (copy_to_user(arg, &buf, sizeof(buf)))
                    return -EFAULT;
            return 0;
    }

    case VIDIOC_QBUF:
    {
            __u32	type = (__u32)arg;
            if (o->noncapturing)
                    return -EPERM;
            if (!dev->stream_buffers_mapped)
                    return -EINVAL;
            if (!capture_queuebuffer(dev, type))
                    return -EINVAL;
            return 0;
    }

    case VIDIOC_NEXTBUF:
    {
            struct timeval tv;
            if (arg)
            {
                    if (copy_from_user(&tv, arg, sizeof(tv)))
                            return -EFAULT;
                    if (!capture_waitforbuffer(dev, &tv))
                            return -EINVAL;
            }
            else if (!capture_waitforbuffer(dev, NULL))
                    return -EINVAL;			
            return 0;
    }

    case VIDIOC_DQBUF:
    {
            struct video_buffer buf;
            if (o->noncapturing)
                    return -EPERM;
            if (copy_from_user(&buf, arg, sizeof(buf)))
                    return -EFAULT;
            if (!capture_dequeuebuffer(dev, &buf))
                    return -EINVAL;
            if (copy_to_user(arg, &buf, sizeof(buf)))
                    return -EFAULT;
            return 0;
    }

    case VIDIOC_STREAMON:
    {
            __u32	type = (__u32)arg;
            if (o->noncapturing)
                    return -EPERM;
            if (!capture_streamon(dev, type))
                    return -EINVAL;
            return 0;
    }

    case VIDIOC_STREAMOFF:
    {
            __u32	type = (__u32)arg;
            if (o->noncapturing)
                    return -EPERM;
            capture_streamoff(dev, type);
            return 0;
    }
#endif

    /* compression */
    case VIDIOC_G_COMP:
    case VIDIOC_S_COMP:
        return -EINVAL;

        /* display window */
    case VIDIOC_G_WIN:
        {
            struct video_window vw;
            vw.x = dev.requested_window.x;
            vw.y = dev.requested_window.y;
            vw.width = dev.requested_window.width;
            vw.height = dev.requested_window.height;
            vw.chromakey = dev.chromakey;
            vw.clips = 0;
            vw.clipcount = 0;
            if (copy_to_user(arg, &vw, sizeof(vw)))
                return -EINVAL;
        }
        return 0;
    case VIDIOC_S_WIN:
        {
            struct video_window vw;
            int r;
            if (copy_from_user(&vw, arg, sizeof(vw)))
                return -EFAULT;
            if (!dev.video_initialized) {
#ifdef DEBUG
                printk("vidcap: running setup_vidcap from a _S_WIN ioctl\n");
#endif
                r = setup_vidcap();
                if (r)
                    return r;
            }
            if (vw.clipcount)
                return -EINVAL;
            dev.requested_window.x = vw.x;
            dev.requested_window.y = vw.y;
            dev.requested_window.width = vw.width;
            dev.requested_window.height = vw.height;
            dev.chromakey = vw.chromakey;
            if (dev.screen.bytes_per_pixel == 2)
                dev.chromakey &= 0xF8FCF8;
            if (dev.screen.bytes_per_pixel == 1)
                dev.chromakey = (dev.chromakey & 0xFF) |
                    ((dev.chromakey & 0xFF) << 8) |
                    ((dev.chromakey & 0xFF) << 16);
            /* no change back to the caller */
            /* if we are active, make the change */
            if (dev.displaying) {
                /* adjust the window to the display;
                 this sets state.realized_window
                 if the window is near the edge,
                 state.requested_window != state.realized_window */
                IGS_AdjustWindow(&dev);
                IGS_SetOverlayFormat (dev.uses_chromakey ? COLORKEY: WINDOWKEY);
                IGS_SetRGBKeyColor (dev.chromakey);
            }

        }
        return 0;

    case VIDIOC_G_FBUF:
    case VIDIOC_S_FBUF:
        {
            struct video_framebuffer vf;
            int r;
            if (cmd==VIDIOC_S_FBUF) {
                if (copy_from_user(&vf, arg, sizeof(vf)))
                    return -EFAULT;

                /* we only take one piece of info from this call */
                dev.screen.offset_vidcap_mem = (unsigned long)vf.base[0] & 0x00FFFFFF;
printk("vidcap memory buffer set %p\n",vf.base[0]);
                dev.mem_buf_set = TRUE;
                dev.uses_chromakey = ((vf.flags & FBUF_FLAG_CHROMAKEY)!=0);
                /* if the user has never called VIDEOC_S_CAP, assume the settings given here */
                if (!dev.capture_format_set) {
                    dev.capture_format.width = vf.fmt.width;
                    dev.capture_format.height = vf.fmt.height;
                }
                /* the caller may be trying to modify the current screen
                 parameters; we re-examine the hardware just in case */
                IGS_GetCurrentScreen(&dev);
                /* otherwise, we don't use this ioctl */
            }
            vf.base[0] = (void*) (dev.screen.offset_vidcap_mem | 0xD0000000);
            vf.flags = (dev.uses_chromakey ? FBUF_FLAG_CHROMAKEY : 0) |
                FBUF_FLAG_OVERLAY;
            vf.capability = FBUF_CAP_CHROMAKEY | FBUF_CAP_SCALEUP;
            vf.fmt.width = dev.capture_format.width;
            vf.fmt.height = dev.capture_format.height;
            vf.fmt.pixelformat = PIX_FMT_YUYV;
            vf.fmt.depth = 16;
            vf.fmt.flags = 0;
            vf.fmt.bytesperline = dev.capture_format.width * 2;
            vf.fmt.sizeimage = dev.capture_format.width * dev.capture_format.height * 2;
            if (copy_to_user(arg, &vf, sizeof(vf)))
                return -EFAULT;
        }
        return 0;
        
        /* show window */
    case VIDIOC_PREVIEW:
        {
            /* if the user passed a 1 or a zero, use it;
             otherwise, assume it is a pointer */
            int i,r;
            if (arg==0) i=0;
            else if (arg==(void*)1) i=1;
            else {
                if (copy_from_user(&i, arg, 1))
                    return -EFAULT;
            }
            if (i && !dev.video_initialized) {
                r = setup_vidcap();
                if (r)
                    return r;
            }
#ifdef DEBUG
            printk("VIDIOC_PREVIEW: turn %s; currently capturing %d, displaying %d, video_init %d\n",
                   i ? "on" : "off",
                   dev.capturing, dev.displaying, dev.video_initialized);
#endif
            if (i) {
                if (!dev.capturing) {
                    struct video_format p;
#ifdef DEBUG
                    printk("VIDIOC_PREVIEW called before VIDIOC_S_FMT or _S_WIN;\n"
                           "    display %d/%d size used for capture\n",
                           dev.requested_window.width,dev.requested_window.height);
#endif
                    p.width = dev.requested_window.width;
                    p.height = dev.requested_window.height;
                    p.pixelformat = PIX_FMT_YUYV;
                    /* update dev, set the hardware and turn on capture */
                    i = capture_new_format(&p,&dev.capture_param);
                    if (i) return i;
                }
                /* adjust the window to the display;
                 this sets state.realized_window
                 if the window is near the edge,
                 state.requested_window != state.realized_window */
                /* similar setup activity is triggered in VIDIOC_S_WIN */
                IGS_AdjustWindow(&dev);
                IGS_SetCaptureFormat (PIX_FMT_YUYV);
                IGS_SetOverlayFormat (dev.uses_chromakey ? COLORKEY: WINDOWKEY);
                IGS_SetRGBKeyColor (dev.chromakey);
                IGS_SetFreezeVidAsync(OFF);
                
                if (!dev.displaying)
                    IGS_WindowOn();
                dev.displaying = TRUE;
            } else {
                IGS_WindowOff();
                dev.displaying = FALSE;
            }
        }
        break;

    case VIDIOC_G_PERF:
        {
            /* number of frames and bytes captured; at this point, returns 0 */
            if (copy_to_user(arg, &dev.perform, sizeof(dev.perform)))
                    return -EFAULT;
            return 0;
    }

    case VIDIOC_G_PARM:
    {
        //            if (copy_to_user(arg, &dev->capture, sizeof(dev->capture)))
        if (copy_to_user(arg, &dev.capture_param, sizeof(dev.capture_param)))
                    return -EFAULT;
        return 0;
    }

    case VIDIOC_S_PARM:
    {
        struct video_parm vp;
        int r;
        if (!dev.video_initialized) {
            r = setup_vidcap();
            if (r)
                return r;
        }

        if (copy_from_user(&vp, arg, sizeof(vp)))
            return -EFAULT;
        if (vp.input < 0 || vp.input >= 1)
            return -EINVAL;
        if (vp.capturemode & ~dev.capture_param.capability)
            return -EINVAL;
        if (vp.timeperframe < 10000)
            return -EINVAL;

        dev.capture_param = vp;

        SetupAndCapture(&dev.capture_format, &dev.capture_param);
        return 0;
    }

    /* not yet implemented */
    case VIDIOC_G_STD:
    case VIDIOC_S_STD:
    case VIDIOC_ENUMSTD:
        return -EINVAL;

    case VIDIOC_ENUMINPUT:
    {
        struct video_input vi;
        if (copy_from_user(&vi, arg, sizeof(vi)))
            return -EFAULT;
        if (vi.index >= 1)
            return -EINVAL;
        return 0;
    }

    case VIDIOC_QUERYCTRL:
    {
        struct video_queryctrl qc;
        int i;
        if (copy_from_user(&qc, arg, sizeof(qc)))
            return -EFAULT;
        for (i=0; i<4; i++)
            if (video_control[i].id == qc.id)
                break;
        if (i == 4)
            return -EINVAL;
        if (copy_to_user(arg, video_control + i, sizeof(qc)))
            return -EFAULT;
        return 0;
    }

    case VIDIOC_G_CTRL:
    {
        struct video_control	vc;
        int			i;
        if (copy_from_user(&vc, arg, sizeof(vc)))
            return -EFAULT;
        for (i=0; i<4; i++)
            if (video_control[i].id == vc.id)
                break;
        if (i == 4)
            return -EINVAL;
        vc.value = dev.control[i];
        if (copy_to_user(arg, &vc, sizeof(vc)))
            return -EFAULT;
        return 0;
    }

    case VIDIOC_S_CTRL:
    {
        struct video_control vc;
        int	i,r;
        if (copy_from_user(&vc, arg, sizeof(vc)))
            return -EFAULT;
        for (i=0; i<4; i++)
            if (video_control[i].id == vc.id)
                break;
        if (i == 4)
            return -EINVAL;
        if (!dev.video_initialized) {
            r = setup_vidcap();
            if (r)
                return r;
        }
        dev.control[i] = vc.value;
        Set7111Params(&dev);
        return 0;
    }

    case VIDIOC_G_TUNER:
    case VIDIOC_S_TUNER:
    case VIDIOC_G_FREQ:
    case VIDIOC_S_FREQ:
    case VIDIOC_G_AUDIO:
    case VIDIOC_S_AUDIO:
        return -EINVAL;

    default:
            return -ENOIOCTLCMD;



            
    case VIDIOC_COREL_FREEZE:
        //IGS_FreezeVideo ();
        if (!IGS_IsGraphics())
            return -EIO;
        IGS_SetFreezeVid(ON);
        break;

    case VIDIOC_COREL_UNFREEZE:
        //IGS_UnfreezeVideo ();
        if (!IGS_IsGraphics())
            return -EIO;
        IGS_SetFreezeVid(OFF);
        break;

    case VIDIOC_COREL_S_SCREENOFFSET:
        {
            struct vidcap_screen_params vsp;
            if (copy_from_user(&vsp, arg, sizeof(vsp)))
                return -EFAULT;
            if (vsp.x_ext<0 || vsp.x_ext>10000 ||
                vsp.y_ext<0 || vsp.y_ext>10000 ||
                vsp.x_orig<0 || vsp.x_orig>vsp.x_ext ||
                vsp.y_orig<0 || vsp.y_orig>vsp.y_ext)
                return -EINVAL;
            dev.screen.x_ext = vsp.x_ext;
            dev.screen.y_ext = vsp.y_ext;
            dev.screen.x_orig = vsp.x_orig;
            dev.screen.y_orig = vsp.y_orig;
            printk("private ioctl: new orig %d %d\n",vsp.x_orig,vsp.y_orig);
            /* only do something if a window exists */
            if (dev.displaying)
                IGS_AdjustWindow(&dev);
        }
        break;
    case VIDIOC_COREL_G_SCREENOFFSET:
        if (copy_to_user(arg, &dev.screen, sizeof(dev.screen)))
            return -EFAULT;
        break;
        
#if 0 // these versions of WindowOn/WindowOff wait for the vga sync; why?
	case CMD_HIDE:
	  {
		while (!(inb_p (0x3da) & 0x08));
		Out_Video_Reg_M (DISP_CTL_I, 0x00, 0x7f);
          }
          state.displaying = FALSE;
	  break;

	case CMD_SHOW:
	  {
		while (!(inb_p (0x3da) & 0x08));
		Out_Video_Reg_M (DISP_CTL_I, 0x80, 0x7f);
          }
          state.displaying = TRUE;
          break;
          
#endif
  } // switch
  return 0;
}

/* come here when the last close() happens */
void
release_video ()
{
    if (dev.displaying)
        IGS_WindowOff ();
    if (dev.capturing)
        IGS_CaptureOff();
    dev.displaying = dev.capturing = FALSE;
    /* turn off the 7111 capture chip */
    set_CPLD(4,4);
    /* restore the FIFO values */
    IGS_VideoCleanUp ();
    dev.video_initialized = FALSE;
}

/**********************************************************************
 ******************* module & insmod code *****************************
 **********************************************************************/
 

#ifndef MODULE
long
vidcap_init (long mem_start, long mem_emd)
{
  if (vidcap_major = register_chrdev (0, vidcap_name, &vidcap_fops))
	printk ("vidcap_init: Unable to get major for Video Capture Device\n");
  else
      printk ("vidcap_init: vidcap_major = %d\n", vidcap_major);

  /* setup the color lookup tables */
  InitXlatTab();

  return mem_start;
}
#else
int
init_module (void)
{
  if (register_chrdev (vidcap_major = VIDEO_MAJOR, vidcap_name, &vidcap_fops) == -EBUSY)
  {
# ifdef DEBUG
      printk ("vidcap_init: Unable to get major for Video Capture Device\n");
# endif
      return -EIO;
  }
# ifdef DEBUG
  else
      printk ("vidcap_init: vidcap_major = %d\n", vidcap_major);
# endif
  InitXlatTab();

  return 0;
}

void
cleanup_module (void)
{
    /* if (state.capturing) ?? do we trust the state? */
  release_video ();
  unregister_chrdev (vidcap_major, vidcap_name);
}
#endif



/***************** memory management functions **********************/
#ifdef notyet
static int
mmap_request_buffers(struct _state* state,
                     struct video_requestbuffers *req)
{
	int	i;

	if (state->stream_buffers_mapped)
		return 0;/* can't make requests if buffers are mapped */
	if (req->count > MAX_CAPTURE_BUFFERS)
		req->count = MAX_CAPTURE_BUFFERS;
	req->type = BUF_TYPE_CAPTURE; /* only kind I know */

printk("wnv: Granting %d buffers\n",req->count);

	/*  Now initialize the buffer structures. Don't allocate the */
	/*  buffers until they're mapped. */
	for (i = 0; i < req->count; ++i)
	{
		dev->stream_buf[i].requested = 1;
		dev->stream_buf[i].vidbuf.index = i;
		dev->stream_buf[i].vidbuf.type = req->type;
		dev->stream_buf[i].vidbuf.offset = 4*i;/* anything unique */
		dev->stream_buf[i].vidbuf.length = dev->capture_buffer_size;
		dev->stream_buf[i].vidbuf.bytesused = 0;
		dev->stream_buf[i].vidbuf.timestamp.tv_sec = 0;
		dev->stream_buf[i].vidbuf.timestamp.tv_usec = 0;
		dev->stream_buf[i].vidbuf.flags = 0;
	}
	for (i = req->count; i < MAX_CAPTURE_BUFFERS; ++i)
		dev->stream_buf[i].requested = 0;

	return 1;
}

static void
mmap_vma_open(struct vm_area_struct *vma)
{
//	struct videum_device *dev =
//		videum_device_from_file(vma->vm_file);
//	if (dev == NULL)
//		return;
	//debug_msg("wnv: vma_open called\n");
	state->stream_buffers_mapped = 1;
	//MOD_INC_USE_COUNT;
}

static void
mmap_vma_close(struct vm_area_struct *vma)
{
#if 0
	struct videum_device *dev =
		videum_device_from_file(vma->vm_file);
	struct stream_buffer *buf =
            mmap_stream_buffer_from_offset(dev, vma->vm_offset);

    
	if (state->streaming)
	{
		info_msg("wnv: Warning- munmap() called while streaming\n");
		capture_streamoff(dev, buf->vidbuf.type);
		v4l2_q_yank_node(&dev->stream_q_capture, &buf->qnode);
		v4l2_q_yank_node(&dev->stream_q_done, &buf->qnode);
        }
#endif
        struct stream_buffer* buf = state->buffer;

	if (buf->vaddress != NULL)
		vfree(buf->vaddress);
	buf->vaddress = NULL;
	if (buf->dma_list)
		free_page((unsigned long)buf->dma_list);
	buf->dma_list = NULL;
	buf->vidbuf.flags = 0;
//debug_msg("wnv: Buffer %d deallocated\n",(int)vma->vm_offset);

	if (dev->stream_buffers_mapped > 0)
		--dev->stream_buffers_mapped;
	//MOD_DEC_USE_COUNT;
}

static unsigned long
mmap_vma_nopage(struct vm_area_struct *vma,
		unsigned long address, int write)
{
	struct videum_device	*dev;
	struct stream_buffer	*buf;
	unsigned long		offset_into_buffer;
	unsigned long		page;

	dev = videum_device_from_file(vma->vm_file);
	if (dev == NULL)
		return 0;
	buf = mmap_stream_buffer_from_offset(dev, vma->vm_offset);
	if (buf == NULL)
		return 0;
	offset_into_buffer = address - vma->vm_start;
	if (offset_into_buffer >= dev->capture_buffer_size)
		return 0;
	page = v4l2_vmalloc_to_page(buf->vaddress + offset_into_buffer);
	if (page == 0)
		return 0;

	atomic_inc(&mem_map[MAP_NR(page)].count);
	return page;
}

static struct vm_operations_struct videum_vma_operations =
{
	mmap_vma_open, mmap_vma_close, NULL, NULL, NULL, NULL,
	mmap_vma_nopage,
};


#endif

int
Set7111Params(struct nw_cap_device* pDev) {
    sendOneByte(P7111Port, 10, pDev->control[CONTROL_BRIGHTNESS]);
    sendOneByte(P7111Port, 11, pDev->control[CONTROL_CONTRAST]);
    sendOneByte(P7111Port, 12, pDev->control[CONTROL_SATURATION]);
    sendOneByte(P7111Port, 13, pDev->control[CONTROL_HUE]);
    return 0;
}
